import ast
from pyproteum.moperators.myoperator import *
import copy


class Oeap(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
		ParentSetter().visit(self.original)


	def clone_node(self, new, old):
		new.targets = [old.target]
		new.value = old.value
		ast.copy_location(new,old)

	def visit_AugAssign(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

		new_node = ast.Assign()
		self.clone_node(new_node, node)
		if idx is not None:
			lst = getattr(p, fld)
			lst[idx] = new_node
		else:
			setattr(p,fld, new_node)
		self.salva_muta(new_node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
		self.seq += 1

		if idx is not None: 
			lst = getattr(p, fld)
			lst[idx] = node
		else:
			setattr(p,fld, node)		

		self.generic_visit(node)
		return node


use = [ast.Add(),
ast.Sub(),
ast.Mult(),
ast.Div(),
ast.FloorDiv(),
ast.Mod(),
ast.Pow(),
ast.LShift(),
ast.RShift(),
ast.BitOr(),
ast.BitXor(),
ast.BitAnd(),
]

class Oepa(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
		self.scalar_set = set()
		ParentSetter().visit(self.original)


	def clone_node(self, new, old, op):
		new.target = old.targets[0]
		new.value = old.value
		new.op = op
		ast.copy_location(new,old)

	def visit_FunctionDef(self, node):
		old_set = set(self.scalar_set)
		super().visit_FunctionDef(node)
		self.scalar_set = old_set
		return node

	def visit_Name(self, node):
		if isinstance(node.ctx, ast.Store) :
			self.scalar_set.add(node.id)
		self.generic_visit(node)
		return node

	def visit_Assign(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

		# não é possível mudar a = b = 10
		if len(node.targets) == 1  and isinstance(node.targets[0], (ast.Attribute,ast.Name) ):
			# só faz a mutação se a variável já foi definida
			for op in use:
				new_node = ast.AugAssign()
				self.clone_node(new_node, node, op)
				if idx is not None:
					lst = getattr(p, fld)
					lst[idx] = new_node
				else:
					setattr(p,fld, new_node)
				self.salva_muta(new_node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1

			if idx is not None: 
				lst = getattr(p, fld)
				lst[idx] = node
			else:
				setattr(p,fld, node)		

		self.generic_visit(node)
		return node



import ast,sys
import difflib

class PrintVisit(ast.NodeVisitor):

	def __init__(self, tree, debug=False):
		self.text = []  # lista de linhas do código final
		self.cur_line = 0
		self.cur_col = 0
		self.debug = debug
		self.closed = False
		self.docstrings = set()
		self.tree = tree
		self.pre_visit(tree)

	def go_visit(self):
		self.visit(self.tree)

	def pre_visit(self, tree):
		for node in ast.walk(tree):
			try:
				if ast.get_docstring(node):
					fn = node.body[0]
					self.docstrings.add(id(fn))
			except: 
				continue
		#print(self.docstrings)

	def show_field(self, field):
		print('Type: ', type(field))
		if isinstance(field,ast.AST):
			print('\t\tunparse: ', ast.unparse(field))
		elif isinstance(field, list):
			if field != []:
				self.show_field(field[0])
				self.show_field(field[1:])
		else:
			print('\t\t', field)

	def show_atributes(self, node):
		if not self.debug:
			return
		if hasattr(node, 'lineno'):
			print('\nNo: {} \tlineno: {} \tcol_offset: {}'.format(type(node),node.lineno, node.col_offset))
		else:
			print('\nNo: {} '.format(type(node)))
		
		print('*******************' , self.cur_col)

		for name,field  in ast.iter_fields(node):
			l = None 
			c = None			
			if hasattr(field, 'lineno'):
				l = field.lineno
				c = field.col_offset
			print('\t{} \tlineno: {} \tcol_offset: {}'.format(name, l, c))
			self.show_field(field)

	def add_text(self, text, line, col):
		if len(text) == 0:
			return
		if not isinstance(text, list):
			text = text.split('\n')
		#col += self.cur_col

		while self.cur_line < line:
			self.text.append('')
			self.cur_line += 1
		else:
			if line < 0:
				self.text.append('')
				self.cur_line += 1

		buffer = self.text[-1]
		buffer += (col-len(buffer)) * ' ' + text[0]

		self.text[-1] = buffer
		self.add_text(text[1:], line+1, 0)

	def gen_visit(self, nodes, col):
		if isinstance(nodes, list):
			for node in nodes:
				self.visit(node)
		elif nodes:
			self.visit(nodes)

	def check_line(self, node):
		if self.closed and node.lineno == self.cur_line:
			self.text[-1] += '; '
		
	def close(self):
		self.closed = True

	def open(self):
		self.closed = False

	

########## No: <class 'ast.FunctionDef'> 	lineno: 11 	col_offset: 0
# 	name 	lineno: None 	col_offset: None
# Type:  <class 'str'>
# 		 exemplo_funcao
# 	args 	lineno: None 	col_offset: None
# Type:  <class 'ast.arguments'>
# 		unparse:  a: int, b: str
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Return'>
# 		unparse:  return a + b
# Type:  <class 'list'>
# 	decorator_list 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Name'>
# 		unparse:  aaaaa
# Type:  <class 'list'>
# Type:  <class 'ast.Name'>
# 		unparse:  bbbbb
# Type:  <class 'list'>
# 	returns 	lineno: 11 	col_offset: 37
# Type:  <class 'ast.Name'>
# 		unparse:  int
# 	type_comment 	lineno: None 	col_offset: None
# Type:  <class 'NoneType'>
# 		 None
# 	type_params 	lineno: None 	col_offset: None
# Type:  <class 'list'>

	def visit_FunctionDef(self, node):
		self.show_atributes(node)
		self.check_line(node) 
		# decorator list comes before function decl
		for name in node.decorator_list:
			self.add_text(f'@{ast.unparse(name)} ', name.lineno, node.col_offset)

		#function name and arguments and returns
		ret = f' -> {ast.unparse(node.returns)}' if node.returns else ''
		self.add_text(f'def {node.name} ({ast.unparse(node.args)}) {ret}:', node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)

	def visit_AsyncFunctionDef(self, node):
		self.show_atributes(node)
		self.check_line(node)
		# decorator list comes before function decl
		for name in node.decorator_list:
			self.add_text(f'@{ast.unparse(name)} ', name.lineno, node.col_offset)

		#function name and arguments and returns
		ret = f' -> {ast.unparse(node.returns)}' if node.returns else ''
		self.add_text(f'async def {node.name} ({ast.unparse(node.args)}) {ret}:', node.lineno, node.col_offset)		
		self.open()
		
		self.gen_visit(node.body, node.col_offset)



# No: <class 'ast.ClassDef'> 	lineno: 19 	col_offset: 0
# 	name 	lineno: None 	col_offset: None
# Type:  <class 'str'>
# 		 ExemploClasse
# 	bases 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Attribute'>
# 		unparse:  ast.AST
# Type:  <class 'list'>
# 	keywords 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.keyword'>
# 		unparse:  metaclass=Meta
# Type:  <class 'list'>
# Type:  <class 'ast.keyword'>
# 		unparse:  abstract=True
# Type:  <class 'list'>
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.FunctionDef'>
# 		unparse:  def metodo(self):
#	 pass
# Type:  <class 'list'>
# 	decorator_list 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# 	type_params 	lineno: None 	col_offset: None
# Type:  <class 'list'>

	def visit_ClassDef(self, node):
		self.show_atributes(node)

		self.check_line(node)
		# decorator list comes before function decl
		for name in node.decorator_list:
			self.add_text(f'@{ast.unparse(name)}', name.lineno, node.col_offset)

		line = f'class {node.name} '
		if node.bases or node.keywords:
			line += '('
			sep = ''
			for base in node.bases:
				line += sep + ast.unparse(base)
				sep = ','
			for keyword in node.keywords:
				line += sep + ast.unparse(keyword)
				sep = ','
			line += ')'
		line += ':'
		self.add_text(line, node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)


####### No: <class 'ast.If'> 	lineno: 65 	col_offset: 0
# 	test 	lineno: 65 	col_offset: 3
# Type:  <class 'ast.Compare'>
# 		unparse:  x > 5
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  print('maior que 5')
# Type:  <class 'list'>
# 	orelse 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.If'>
# 		unparse:  if x == 5:
#	 print('igual a 5')
# else:
#	 print('menor ou igual a 5')
# Type:  <class 'list'>

	def visit_If(self, node, elif_flag=False):
		self.check_line(node)
		line = 'elif ' if elif_flag else 'if '
		line += ast.unparse(node.test) + ':'
		self.add_text(line, node.lineno, node.col_offset)

		self.show_atributes(node)
		self.open()
		self.gen_visit(node.body, node.col_offset)

		if node.orelse:
			self.open()
			if len(node.orelse) == 1:
				next_if = node.orelse[0]
				# se o comando do else é um só e é um if, então trata-se de um elif
				if isinstance(next_if, ast.If) and next_if.col_offset == node.col_offset:
					self.cur_col = node.col_offset
					self.visit_If(next_if,True)
				else:
					self.add_text('else:', -1, node.col_offset)
					self.visit(next_if)
			else:
				self.add_text('else:', -1, node.col_offset)
				self.gen_visit(node.orelse, node.col_offset)




################# No: <class 'ast.For'> 	lineno: 81 	col_offset: 0
# 	target 	lineno: 81 	col_offset: 4
# Type:  <class 'ast.Tuple'>
# 		unparse:  (i, j, k)
# 	iter 	lineno: 81 	col_offset: 13
# Type:  <class 'ast.Call'>
# 		unparse:  range(2)
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  print('for', i)
# Type:  <class 'list'>
# 	orelse 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  print('fim do for')
# Type:  <class 'list'>
# 	type_comment 	lineno: None 	col_offset: None
# Type:  <class 'NoneType'>

	def visit_For(self, node):
		self.check_line(node)
		self.show_atributes(node)

		line = f'for {ast.unparse(node.target)} in {ast.unparse(node.iter)}:'
		self.add_text(line, node.lineno, node.col_offset)
		self.open()

		self.gen_visit(node.body, node.col_offset)
		if node.orelse:
			self.add_text('else:', -1, node.col_offset)
			self.open()
			self.gen_visit(node.orelse, node.col_offset)


	def visit_AsyncFor(self, node):
		self.check_line(node)
		self.show_atributes(node)

		line = f'async for {ast.unparse(node.target)} in {ast.unparse(node.iter)}:'
		self.add_text(line, node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)
		if node.orelse:
			self.add_text('else:', -1, node.col_offset)
			self.open()
			self.gen_visit(node.orelse, node.col_offset)


################### No: <class 'ast.While'> 	lineno: 76 	col_offset: 0
# ******************* 0
# 	test 	lineno: 76 	col_offset: 6
# Type:  <class 'ast.Compare'>
# 		unparse:  i < 3
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  print(i)
# Type:  <class 'list'>
# Type:  <class 'ast.AugAssign'>
# 		unparse:  i += 1
# Type:  <class 'list'>
# 	orelse 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  print('fim do while')
# Type:  <class 'list'>

	def visit_While(self, node):
		self.check_line(node)
		self.show_atributes(node)

		line = f'while {ast.unparse(node.test)} :'
		self.add_text(line, node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)
		if node.orelse:
			self.open()
			self.add_text('else:', -1, node.col_offset)
			self.gen_visit(node.orelse, node.col_offset)


# ############No: <class 'ast.With'> 	lineno: 109 	col_offset: 0
# ******************* 0
# 	items 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.withitem'>
# 		unparse:  open('exemplo.txt', 'w') as f
# Type:  <class 'list'>
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Expr'>
# 		unparse:  f.write('teste')
# Type:  <class 'list'>
# 	type_comment 	lineno: None 	col_offset: None
# Type:  <class 'NoneType'>
# 		 None

	def visit_With(self, node):
		self.show_atributes(node)
		self.check_line(node)
		with_item = ''
		sep = ''
		for item in node.items:
			with_item += f'{sep}{ast.unparse(item)}'
			sep = ','  
		self.add_text(f'with {with_item}:', node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)


	def visit_AsyncWith(self, node):
		self.show_atributes(node)
		self.check_line(node)
		with_item = ''
		sep = ''
		for item in node.items:
			with_item += f'{sep}{ast.unparse(item)}'
			sep = ','  
		self.add_text(f'async with {with_item}:', node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)

		

################### No: <class 'ast.Try'> 	lineno: 44 	col_offset: 0
# ******************* 0
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Raise'>
# 		unparse:  raise ValueError('Erro')
# Type:  <class 'list'>
# 	handlers 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.ExceptHandler'>
# 		unparse:  except ValueError as ex:
#	 pass
# Type:  <class 'list'>
# 	orelse 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# 	finalbody 	lineno: None 	col_offset: None
# Type:  <class 'list'>

	def visit_Try(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text('try:', node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)
		if node.handlers:
			self.gen_visit(node.handlers, node.col_offset)
		if node.orelse:
			self.add_text('else:', -1, node.col_offset)
			self.open()
			self.gen_visit(node.orelse, node.col_offset)
		if node.finalbody:
			self.add_text('finally:', -1, node.col_offset)
			self.open()
			self.gen_visit(node.finalbody, node.col_offset)

################# No: <class 'ast.ExceptHandler'> 	lineno: 99 	col_offset: 0
# ******************* 0
# 	type 	lineno: 99 	col_offset: 7
# Type:  <class 'ast.Name'>
# 		unparse:  ZeroDivisionError
# 	name 	lineno: None 	col_offset: None
# Type:  <class 'NoneType'>
# 		 None
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Assign'>
# 		unparse:  x = 0
# Type:  <class 'list'>

	def visit_ExceptHandler(self,  node):
		self.show_atributes(node)
		self.check_line(node)
		name = f'as {node.name}' if node.name else '' 
		tipo = ast.unparse(node.type) if node.type else ''
		self.add_text(f'except {tipo} {name}:', node.lineno, node.col_offset)
		self.open()
		self.gen_visit(node.body, node.col_offset)		

######### No: <class 'ast.Assign'> 	lineno: 28 	col_offset: 0
# ******************* 0
# 	targets 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Name'>
# 		unparse:  a
# Type:  <class 'list'>
# 	value 	lineno: 28 	col_offset: 4
# Type:  <class 'ast.List'>
# 		unparse:  [1, 2, 3]
# 	type_comment 	lineno: None 	col_offset: None
# Type:  <class 'NoneType'>
# 		 None

	def visit_Assign(self, node):
		self.check_line(node)
		self.show_atributes(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

############### No: <class 'ast.AugAssign'> 	lineno: 35 	col_offset: 0
# ******************* 0
# 	target 	lineno: 35 	col_offset: 0
# Type:  <class 'ast.Name'>
# 		unparse:  x
# 	op 	lineno: None 	col_offset: None
# Type:  <class 'ast.Add'>
# 		unparse:  
# 	value 	lineno: 35 	col_offset: 5
# Type:  <class 'ast.Constant'>
# 		unparse:  5

	def visit_AugAssign(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()
		
	def visit_AnnAssign(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

##### No: <class 'ast.Expr'> 	lineno: 16 	col_offset: 4
# ******************* 0
# 	value 	lineno: 16 	col_offset: 4
# Type:  <class 'ast.Await'>
# 		unparse:  await exemplo_corrotina()

	def visit_Expr(self, node):
		self.show_atributes(node)
		if id(node) in self.docstrings:
			self.visit_docstring(node)
		else:
			self.check_line(node)
			self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_docstring(self, node):
		#print('Docstring: ', node.value.value, len(node.value.value))
		self.add_text(f"'''{node.value.value}'''", node.lineno, node.col_offset)

	def visit_Return(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Raise(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Assert(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Delete(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()


	def visit_Global(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Nonlocal(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Break(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Continue(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_Pass(self, node):
		self.show_atributes(node)
		self.check_line(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()


################# No: <class 'ast.Match'> 	lineno: 97 	col_offset: 0
# ******************* 0
# 	subject 	lineno: 97 	col_offset: 6
# Type:  <class 'ast.Name'>
# 		unparse:  x
# 	cases 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.match_case'>
# 		unparse:  case 1 if j > 10:
#     pass
# Type:  <class 'list'>
# Type:  <class 'ast.match_case'>
# 		unparse:  case _:
#     pass
# Type:  <class 'list'>

################# No: <class 'ast.match_case'> 
# ******************* 0
# 	pattern 	lineno: 98 	col_offset: 6
# Type:  <class 'ast.MatchValue'>
# 		unparse:  1
# 	guard 	lineno: 98 	col_offset: 11
# Type:  <class 'ast.Compare'>
# 		unparse:  j > 10
# 	body 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.Pass'>
# 		unparse:  pass
# Type:  <class 'list'>

	def visit_Match(self, node):
		self.show_atributes(node)
		self.add_text(f'match {ast.unparse(node.subject)}:', node.lineno, node.col_offset)
		for cs in node.cases: 
			gd = ''
			if cs.guard:
				gd = f'if {ast.unparse(cs.guard)}'
			self.add_text(f'\ncase {ast.unparse(cs.pattern)} {gd}:',-1 ,node.col_offset+4)
			self.open()
			self.gen_visit(cs.body, node.col_offset+4)
		

	def visit_match_case(self, node):
		self.show_atributes(node)
		self.generic_visit(node)	

############### No: <class 'ast.Import'> 	lineno: 111 	col_offset: 0
# ******************* 0
# 	names 	lineno: None 	col_offset: None
# Type:  <class 'list'>
# Type:  <class 'ast.alias'>
# 		unparse:  asyncio
# Type:  <class 'list'>

	def visit_Import(self, node):
		self.check_line(node)
		self.show_atributes(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def visit_ImportFrom(self, node):
		self.check_line(node)
		self.show_atributes(node)
		self.add_text(f'{ast.unparse(node)}', node.lineno, node.col_offset)
		self.close()

	def __str__(self):
		s = ''
		for r in self.text:
			s += r+'\n'
		return s


def compare_ast(node1, node2, path="root"):
	if isinstance(node1, ast.AST):
		print(path, ast.unparse(node1), ast.unparse(node2))
	if type(node1) != type(node2):
		print(f"Tipo diferente em {path}: {type(node1).__name__} vs {type(node2).__name__}")
		print(node1, node2)		
		return

	if isinstance(node1, ast.AST):
		for field in node1._fields:
			val1 = getattr(node1, field, None)
			val2 = getattr(node2, field, None)
			compare_ast(val1, val2, path + f".{field}")
	elif isinstance(node1, list):
		if len(node1) != len(node2):
			print(f"Listas de tamanhos diferentes em {path}: {len(node1)} vs {len(node2)}")
			print(node1)
			print(node2)
			
		for i, (item1, item2) in enumerate(zip(node1, node2)):
			compare_ast(item1, item2, path + f"[{i}]")
	else:
		if node1 != node2:
			print(f"Diferença em {path}: {node1!r} vs {node2!r}")
			if node1.lineno:
				print(node1.lineno)

if __name__ == "__main__":
	source = open(sys.argv[1]).read()
	tree = ast.parse(source)
	#print(ast.dump(tree))
	pv = PrintVisit(tree, debug=True)
	pv.go_visit()
	print(pv)

	tree2 = ast.parse(str(pv))

	d1 = ast.dump(tree)
	d2 = ast.dump(tree2)

	if d1 == d2:
		print('iguais')
	else:
		compare_ast(tree, tree2)
		with open('../erro.py', 'w') as f:
			f.write(str(pv))



import ast
from pyproteum.moperators.myoperator import *


class Crcr(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
	
	def visit_Constant(self, node):
		old_value = node.value

		if not id(node) in self.docstrings:

			for new_value in self.get_required(node.value):
				if new_value == old_value:
					continue
				node.value = new_value
				self.salva_muta(node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1
			node.value = old_value
		return self.generic_visit(node)

	def get_required(self, value):
		if isinstance(value, int):
			return [-1, 0, 1]
		if isinstance(value, float):
			return [-1.0, 0.0, 1.0, 1E-13, 1E+13]
		if isinstance(value, str):
			return ['',' ']		
		if isinstance(value, bytes):
			return [b'',b' ']		
		return []



from peewee import *
from datetime import datetime

db = Proxy()

class BaseModel(Model):
	class Meta:
		database = db

class Session(BaseModel):
	filename = CharField(unique=True)
	type = TextField()
	#ast = BlobField()
	criado = DateTimeField(default=datetime.now)

class TestCase(BaseModel):
	filename = CharField(unique=True)
	criado = DateTimeField(default=datetime.now)


class Mutant(BaseModel):
	source = ForeignKeyField(Session, backref='mutants')
	operator = TextField()
	function = TextField()
	func_lineno = IntegerField()
	func_end_lineno = IntegerField()
	lineno = IntegerField()
	col_offset = IntegerField()
	end_lineno = IntegerField()
	end_col_offset = IntegerField()
	seq_number = IntegerField()
	ast = BlobField()
	status = TextField(default='live')
	criado = DateTimeField(default=datetime.now)

	class Meta:
		indexes = (
			(('source', 'lineno', 'col_offset','operator','seq_number'), True),  # combinação única
		)

	def __str__(self):
		s = f'Source: {self.source.filename}\n'
		s += f'Operator: {self.operator}\n'
		s += f'Func Lineno: {self.func_lineno}\n'
		s += f'Func End Lineno: {self.func_end_lineno}\n'
		s += f'Lineno: {self.lineno}\n'
		s += f'End Lineno: {self.end_lineno}\n'
		return s

def criar_modelo(nome_modelo, campos):
	class Meta:
		database = db

	atributos = dict(campos)
	atributos['Meta'] = Meta
	return type(nome_modelo, (Model,), atributos)

#uso: 
# campos = {
#	 'titulo': CharField(),
#	 'ano': IntegerField(),
#	 'autor': CharField(),
# }

# Livro = criar_modelo('Livro', campos, db)
# db.create_tables([Livro])
# Livro.create(titulo='Dom Casmurro', ano=1899, autor='Machado de Assis')
#db.drop_tables([Model1, Model2])

def get_table_model(table_name):
	colunas = db.get_columns(table_name)
	atributos = {
		'_meta': type('Meta', (), {'database': db, 'table_name': table_name})
	}
	for col in colunas:
		atributos[col.name] = Field()
	return type(table_name, (BaseModel,), atributos)

# Livro = get_table_model('Livro')

def criar_modelo_dinamico(nome_tabela):
	colunas = db.get_columns(nome_tabela)
	atributos = {
		'_meta': type('Meta', (), {'database': db, 'table_name': nome_tabela})
	}
	for col in colunas:
		# Usa Field genérico para não errar o tipo
		atributos[col.name] = Field()
	return type(nome_tabela, (BaseModel,), atributos)


import sys
import os, types
from pyproteum.models.models import *
import unittest
import importlib.util

from pyproteum.moperators.myoperator import *
from pyproteum.utiles import red,green,get_test_files_with_discover



def change_dir_connect(d, session_name):
	print('Session: ', session_name)
	try :
		if d:
			os.chdir(d)
			sys.path.insert(0, os.getcwd())


		database = SqliteDatabase(session_name+'.db')
		db.initialize(database)  # aqui o proxy é vinculado ao banco real
		db.pragma('foreign_keys', 1, permanent=True)
	except Exception as ex:
		print(red('Error: test session not found'))
		print(red(ex))
		sys.exit()


def _to_module_name(key: str) -> str:
	# "dir/arquivo.py" -> "dir.arquivo"
	# "dir\\arquivo.py" (Windows) -> "dir.arquivo"
	# "arquivo.py" -> "arquivo"
	# "dir.arquivo" -> "dir.arquivo"
	if key.endswith(".py"):
		key = os.path.splitext(key)[0]
	key = key.replace(os.sep, ".")
	return key

def _ensure_module_hierarchy(module_name: str):
	module_name = _to_module_name(module_name)
	parts = module_name.split(".")
	parent = None
	fullname = ""
	for part in parts[:-1]:
		fullname = f"{fullname}.{part}" if fullname else part
		if fullname in sys.modules:
			parent = sys.modules[fullname]
			continue
		# verifica se é um pacote real no disco
		if importlib.util.find_spec(fullname) is not None:
			# importa o pacote de verdade
			__import__(fullname)
			parent = sys.modules[fullname]
			continue
		# cria pacote "fake" só se não existe nem no sys.modules nem no disco
		pkg = types.ModuleType(fullname)
		pkg.__path__ = []  # pacote vazio
		sys.modules[fullname] = pkg
		if parent is not None:
			setattr(parent, part, pkg)
		parent = pkg
	return module_name, parent, parts
	
def load_modulex(filename):
	nome_modulo = _to_module_name(filename)
	spec = importlib.util.spec_from_file_location(nome_modulo, filename)
	modulo = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(modulo)
	return modulo

def load_module(filename):
	module_name, parent, parts = _ensure_module_hierarchy(filename)

	spec = importlib.util.spec_from_file_location(module_name, filename)
	modulo = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(modulo)

	sys.modules[module_name] = modulo
	if parent is not None:
		setattr(parent, parts[-1], modulo)

	return modulo

def get_test_names(filename):
	module = load_module(filename)
	loader = unittest.TestLoader()
	suite = loader.loadTestsFromModule(module)
	tests = []
	for test_group in suite:
		for test in test_group:
			tests.append(test.id())
	return sorted(tests)

def __add():
	session_name = sys.argv[-1]
	directory = None
	tests = []
	i = 2
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--S':
				i += 1
				tests.append(sys.argv[i])
			case '--discover':
				i += 1
				vnd = get_test_files_with_discover(sys.argv[i], '*.py')
				tests += vnd 
			case _:
				usage()
				return
		i += 1

	
	if len(tests) == 0:
		tests = ['test_'+session_name+'.py']

	print('Test files to add:')
	for s in tests:
		print(f'\t{s}')
	
	change_dir_connect(directory, session_name)

	for test_file in tests:
		try :
			test_names = get_test_names(test_file)
			TestCase.create(filename=test_file)
		except Exception as ex:
			print(red(f'Error: can not insert test file {test_file}'))
			print(red(ex))
			continue

		print(f'\nTests form file {test_file}')
		i = 1
		for t in test_names:
			t = t.split('.')[-1]
			print(f'\t{i}) {t}')
			i += 1


def __list():
	session_name = sys.argv[-1]
	directory = None
	tests = []
	i = 2
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case _:
				usage()
				return
		i += 1

	
	change_dir_connect(directory, session_name)


	k = 1
	nn = len(TestCase)
	for test_file in TestCase:
		test_file = test_file.filename
		try :
			test_names = get_test_names(test_file)
		except Exception as ex:
			print(red(f'Error: can not find test file {test_file}'))
			print(red(ex))
			continue

		print(f'\nTests form file {test_file} ({k})')
		i = 1
		for t in test_names:
			t = t.split('.')[-1]
			print(f'\t{k}:{i}) {t}')
			i += 1
		k += 1


def get_all_test_names():
	k = 1
	all = []
	for test_file in TestCase:
		test_file = test_file.filename
		try :
			test_names = get_test_names(test_file)
			all += test_names
		except Exception as ex:
			print(red(f'Error: can not find test file {test_file}'))
			print(red(ex))
			continue

		k += 1
	return all



def __delete():
	session_name = sys.argv[-1]
	directory = None
	tests = []
	i = 2
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--S':
				i += 1
				tests.append(sys.argv[i])
			case '--discover':
				i += 1
				vnd = get_test_files_with_discover(sys.argv[i])
				tests += vnd 			
			case _:
				usage()
				return
		i += 1

	if len(tests) == 0:
		tests = ['test_'+session_name+'.py']
	
	change_dir_connect(directory, session_name)

	k = 1
	for test_file in tests:
		try:
			tc = TestCase.get(TestCase.filename == test_file)
			tc.delete_instance()
			print(f'Succesfuly removed "{test_file}"')
		except TestCase.DoesNotExist:
			print(red(f'No file "{test_file}" found.'))

def main():
	n = len(sys.argv)-2
	if n < 1:
		usage()
	
	if sys.argv[1] == '--add':
		__add()
		return
	elif sys.argv[1] == '--del':
		__delete()
		return	
	elif sys.argv[1] == '--list':
		__list()
		return 
	else:
		usage()

 
def usage():
	print('Usage:')
	print('tcase --add [--D <directory> ] [--S <test file name> ... ] [--discover <directory name>] <session name>')
	print('tcase --del [--D <directory> ] [--S <test file name> ...]  <session name>')
	print('tcase --list [--D <directory> ]  <session name>')
	sys.exit()


if __name__ == '__main__' :
	main()



import sys
from random import Random
import os
from pyproteum.models.models import *
import pickle
from peewee import IntegrityError
from pyproteum.moperators.myoperator import *
from pyproteum.moperators.oplist import operator_list
from pyproteum.utiles import red,green
from pyproteum.tcase import change_dir_connect

rd = Random()

def main():
	n = len(sys.argv)-2
	if n < 1:
		usage()
	
	if sys.argv[1] == '--create':
		__create()
		return
	else:
		usage()

def find_op(name):
	name = name.lower()
	l = []
	for r in operator_list:
		if r['name'].startswith(name):
			l.append(r['class'])
	if l == []:
		raise ValueError('Opertor not found.')
	return l

def __create():
	session_name = sys.argv[-1]
	directory = None
	i = 2
	ops = {}
	percent = 100
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--all':
				try:
					i += 1
					p = float(sys.argv[i]) 
					if p > 100: p = 100
					if p < 0 : p = 0						
					for r in  find_op(''):
						ops[r] = p
				except:
					usage()
					return			
			case '--seed':
				try:
					i += 1
					seed = int(sys.argv[i])
					rd.seed(seed)
				except:
					usage()
					return				
			case _:
				try:
					i += 1
					p = float(sys.argv[i]) 
					if p > 100: p = 100
					if p < 0 : p = 0					
					for r in  find_op(s[2:]):
						ops[r] = p
				except:
					usage()
					return
		i += 1
	if len(ops) == 0:
		for r in find_op(''):
			ops[r] = 100

	muta_create(directory, session_name, ops)


def usage():
	print('Usage:')
	print('mutagen --create [--D <directory> ] [(--<mutant op> |--all) <percentage>] [--seed <integer number>] <session name>')
	print('\t--<mutant operator> means the name of an operator or a preffix that matches one or many opertors ')
	print('\tFor instance "--s" matches all operators beginning with an "s", case insensitive.')
	print('\t--all means all operators')
	print('\t--seed stablishes the seed for mutant sampling. Using the same seed generates the same mutants.')
	sys.exit()
				
def muta_create(d, session_name, ops):
	change_dir_connect(d, session_name)

	ttotal = 0
	for reg_session in Session:
		mutants = apply_mutations(reg_session.filename, ops)
		print(green(f'Generated {len(mutants)} total mutants for file {reg_session.filename}'))
		ttotal += len(mutants)
	   # print(mutants[-1])
		insert_db(mutants, reg_session)

	print(green(f'Generated {ttotal} mutants for all source files'))


def insert_db(muta, reg_session):
	ignore = None

	with db.atomic():
		for m in muta:
			if m['operator'] == ignore:
				continue
			ignore = None
			try: 
				Mutant.create(
					source = reg_session,
					operator = m['operator'],
					function = m['function'],
					func_lineno = m['func_lineno'],
					func_end_lineno = m['func_end_lineno'],
					lineno = m['lineno'],
					col_offset = m['col_offset'],
					end_lineno = m['end_lineno'],
					end_col_offset = m['end_col_offset'],
					seq_number = m['seq_number'],
					ast = pickle.dumps(m['ast'])
				)

			except IntegrityError as ex:
				print(red(f'Can not insert mutants {m["operator"]}. They probably already are in the test session.'))
				print(red(ex))
				ignore = m['operator']
			except Exception as ex:
				print(red('Can not insert mutants.'))
				print(red(type(ex)), red(ex))
				return


def apply_mutations(filename, oper_list):

	try:
		with open(filename) as f:
			source = f.read()
	except Exception as ex:
		print(red('Error: Can´t read source file'))
		print(red(ex))
		sys.exit()
	
	try:
		tree = ast.parse(source.replace('\t', '    '))
	except Exception as ex:
		print(red(f'Error in file {filename}. Check the errors'))
		print(red(ex))
		sys.exit()

	mutants = []
	for op_class,percentage in oper_list.items():
		op = op_class(tree, filename)
		print(f'Applying {str(op)}')
		op.go_visit()
		sampled = selecionar_percentual(op.mutants, percentage)
		print(f'Generated {len(sampled)} mutants for {str(op)}')
		mutants += sampled

	return mutants
		
def selecionar_percentual(lista, n):
	if n < 0: n = 0
	if n > 100: n = 100
	if lista == [] :
		return []
	quantidade = max(1, int(len(lista) * n / 100)) if n > 0 else 0
	return sorted(rd.sample(lista, quantidade))

if __name__ == '__main__' :
	main()









import tkinter as tk
from tkinter import ttk
from tkinter import font as tkfont
from pyproteum.models.models import Mutant
from pyproteum.print_visit import PrintVisit
import ast
import pickle
from peewee import fn
from pyproteum.moperators.oplist import operator_list,get_descr
import sys, random

# ---------------------- Util ----------------------
def log(msg):
	#print(msg, flush=True)
	pass

def clamp_int(text, default=1):
	try:
		return max(0, int(text))
	except:
		return default

# ---------------------- App ----------------------
class ViewMutantsApp(ttk.Frame):
	def __init__(self, master):
		super().__init__(master, padding=6)
		self.master = master
		self.grid(sticky="nsew")
		self._make_vars()
		self._make_ui()
		self._layout()
		self._bind_events()

	def _make_vars(self):
		self.var_mutant = tk.StringVar(value="1")
		self.var_op_equiv = tk.IntVar(value=0)
		self.var_show_alive = tk.IntVar(value=1)
		self.var_show_dead = tk.IntVar(value=1)
		self.var_show_equiv = tk.IntVar(value=1)
		self.var_status = tk.StringVar()
		self.var_operator = tk.StringVar()
		self.mono = tkfont.Font(family="Courier New", size=10)
		
	def init_fields(self, muta):
		from pyproteum.mutaview import get_source_orig_muta

		try:
			reg_muta = Mutant.get(Mutant.id==muta)
		except:
			return -1
		self.last = muta
		self.var_mutant.set(str(muta))
		self.var_status.set(reg_muta.status)
		self.var_operator.set(f'{reg_muta.operator}: {get_descr(reg_muta.operator)}')
		self.var_op_equiv.set(int(reg_muta.status=='equiv'))
		if reg_muta.status == 'dead':
			self.chk_op_equiv.config(state='disabled')
		else:
			self.chk_op_equiv.config(state='normal')
		self.txt_right.configure(state="normal")
		self.txt_right.delete("1.0", tk.END)
		sample_mutant2, sample_mutant1 = get_source_orig_muta(reg_muta, color=False)
		self.txt_right.insert("1.0", sample_mutant1)
		self.txt_right.configure(state="disabled")

		self.txt_left.configure(state="normal")
		self.txt_left.delete("1.0", tk.END)
		self.txt_left.insert("1.0", sample_mutant2)
		self.txt_left.configure(state="disabled")
		return muta

	def _make_ui(self):
		# Linha 1
		self.lbl_mutant = ttk.Label(self, text="Mutant:")
		self.ent_mutant = ttk.Entry(self, textvariable=self.var_mutant, width=8)
		self.btn_up = ttk.Button(self, text="up", command=self.on_up)
		self.btn_dw = ttk.Button(self, text="down", command=self.on_dw)
		self.chk_op_equiv = ttk.Checkbutton(self, text="Equivalent", variable=self.var_op_equiv, command=self.on_op_equiv)

		self.lbl_type = ttk.Label(self, text="Type to Show:")
		self.frm_types = ttk.Frame(self)

		self.chk_alive = ttk.Checkbutton(self.frm_types, text="Alive", variable=self.var_show_alive, command=self.on_show_alive)
		self.chk_equiv = ttk.Checkbutton(self.frm_types, text="Equivalent", variable=self.var_show_equiv, command=self.on_show_equiv)
		self.chk_dead  = ttk.Checkbutton(self.frm_types, text="Dead", variable=self.var_show_dead, command=self.on_show_dead)
		# Linha 2 (desabilitados)
		self.lbl_status = ttk.Label(self, text="Status:")
		self.ent_status = ttk.Entry(self, textvariable=self.var_status, width=20, state="disabled")
		self.lbl_operator = ttk.Label(self, text="Operator:")
		self.ent_operator = ttk.Entry(self, textvariable=self.var_operator, width=100, state="disabled")

		# Títulos das áreas
		self.lbl_orig_title = ttk.Label(self, text="Original Program", font=("Arial", 11, "bold"))
		self.lbl_mut_title = ttk.Label(self, text="Mutant Program", font=("Arial", 11, "bold"))

		# Textos com rolagem (lado a lado)
		self.frm_left = ttk.Frame(self)
		self.txt_left = tk.Text(self.frm_left, width=70, height=24, font=self.mono, wrap="none")
		self.scrl_left_y = ttk.Scrollbar(self.frm_left, orient="vertical", command=self.txt_left.yview)
		self.scrl_left_x = ttk.Scrollbar(self.frm_left, orient="horizontal", command=self.txt_left.xview)
		self.txt_left.configure(yscrollcommand=self.scrl_left_y.set, xscrollcommand=self.scrl_left_x.set)
		sample_original = ""
		self.txt_left.insert("1.0", sample_original)
		self.txt_left.configure(state="disabled")

		self.frm_right = ttk.Frame(self)
		self.txt_right = tk.Text(self.frm_right, width=70, height=24, font=self.mono, wrap="none")
		self.scrl_right_y = ttk.Scrollbar(self.frm_right, orient="vertical", command=self.txt_right.yview)
		self.scrl_right_x = ttk.Scrollbar(self.frm_right, orient="horizontal", command=self.txt_right.xview)
		self.txt_right.configure(yscrollcommand=self.scrl_right_y.set, xscrollcommand=self.scrl_right_x.set)
		sample_mutant = ""
		self.txt_right.insert("1.0", sample_mutant)
		self.txt_right.configure(state="disabled")

		# Rodapé
		self.btn_ok = ttk.Button(self, text="OK", command=self.master.destroy)

	def _layout(self):
		self.columnconfigure(0, weight=0)
		self.columnconfigure(1, weight=0)
		self.columnconfigure(2, weight=0)
		self.columnconfigure(3, weight=0)
		self.columnconfigure(4, weight=0)
		self.columnconfigure(5, weight=0)

		# Linha 1
		self.lbl_mutant.grid(row=0, column=0, padx=(4,2), pady=(4,2), sticky="w")
		self.ent_mutant.grid(row=0, column=1, padx=(0,6), pady=(4,2), sticky="w")
		self.btn_up.grid(row=0, column=2, padx=(0,2), pady=(4,2), sticky="w")
		self.btn_dw.grid(row=0, column=3, padx=(0,12), pady=(4,2), sticky="w")
		self.chk_op_equiv.grid(row=0, column=4, padx=(0,16), pady=(4,2), sticky="w")

		# o rótulo "Type to Show:"
		self.lbl_type.grid(row=0, column=5, padx=(0,6), pady=(4,2), sticky="w")

		# 🔸 frame que vai conter os 3 checkboxes
		# ele ocupa da coluna 6 até 8, mas como um único widget
		self.frm_types.grid(row=0, column=6, columnspan=3, padx=(0,4), pady=(4,2), sticky="ew")

		# o frame expande
		self.columnconfigure(6, weight=1)
		self.columnconfigure(7, weight=1)
		self.columnconfigure(8, weight=1)

		# 🔸 dentro do frame: 3 colunas iguais
		for c in range(3):
			self.frm_types.columnconfigure(c, weight=1, uniform="types3")

		# cada checkbox em sua coluna, preenchendo horizontalmente
		self.chk_alive.grid(row=0, column=0, sticky="ew", padx=(0,4))
		self.chk_equiv.grid(row=0, column=1, sticky="ew", padx=(0,4))
		self.chk_dead.grid(row=0, column=2, sticky="ew")


		# Linha 2
		self.lbl_status.grid(row=1, column=0, padx=(4,2), pady=(2,6), sticky="w")
		self.ent_status.grid(row=1, column=1, padx=(0,12), pady=(2,6), sticky="w")
		self.lbl_operator.grid(row=1, column=2, padx=(0,2), pady=(2,6), sticky="w")
		self.ent_operator.grid(row=1, column=3, columnspan=4, padx=(0,12), pady=(2,6), sticky="we")

		# Títulos
		self.lbl_orig_title.grid(row=2, column=0, columnspan=5, pady=(2,2), sticky="w")
		self.lbl_mut_title.grid(row=2, column=5, columnspan=6, pady=(2,2), sticky="w")

		# Textos com rolagem
		self.frm_left.grid(row=3, column=0, columnspan=5, sticky="nsew", padx=(0,6), pady=(2,6))
		self.frm_left.rowconfigure(0, weight=1)
		self.frm_left.columnconfigure(0, weight=1)
		self.txt_left.grid(row=0, column=0, sticky="nsew")
		self.scrl_left_y.grid(row=0, column=1, sticky="ns")
		self.scrl_left_x.grid(row=1, column=0, sticky="ew")

		self.frm_right.grid(row=3, column=5, columnspan=6, sticky="nsew", padx=(6,0), pady=(2,6))
		self.frm_right.rowconfigure(0, weight=1)
		self.frm_right.columnconfigure(0, weight=1)
		self.txt_right.grid(row=0, column=0, sticky="nsew")
		self.scrl_right_y.grid(row=0, column=1, sticky="ns")
		self.scrl_right_x.grid(row=1, column=0, sticky="ew")

		# Rodapé
		self.btn_ok.grid(row=4, column=0, columnspan=11, pady=(2,2))

		# Expansão
		self.rowconfigure(3, weight=1)

	def _bind_events(self):
		# Confirma Mutant apenas no Enter
		self.ent_mutant.bind("<Return>", self.on_mutant_enter)

	# ------------------ Handlers ------------------
	def on_mutant_enter(self, event=None):
		text = self.var_mutant.get()
		new_id = clamp_int(text, default=1)
		self.var_mutant.set(str(new_id))
		# mantém o foco no campo após confirmar
		self.ent_mutant.focus_set()
		if self.init_fields(int(self.var_mutant.get())) < 0:
			print('Not set.')
			self.var_mutant.set(str(self.last))

	def on_up(self):
		text = self.var_mutant.get()
		cur = clamp_int(self.var_mutant.get(), default=1) + 1
		while True:
			try:
				reg_muta = Mutant.get(Mutant.id==cur)
			except:
				cur = self.last
				break
			if reg_muta.status == 'live' and bool(self.var_show_alive.get()):
				break
			elif reg_muta.status == 'dead' and bool(self.var_show_dead.get()):
				break
			elif reg_muta.status == 'equiv' and bool(self.var_show_equiv.get()):
				break
			else:
				cur += 1
		self.var_mutant.set(str(cur))
		self.init_fields(cur)
		log(f"[Mutant] up -> {cur}")

	def on_dw(self):
		text = self.var_mutant.get()
		cur = clamp_int(self.var_mutant.get(), default=1) - 1
		while True:
			try:
				reg_muta = Mutant.get(Mutant.id==cur)
			except:
				cur = self.last
				break
			if reg_muta.status == 'live' and bool(self.var_show_alive.get()):
				break
			elif reg_muta.status == 'dead' and bool(self.var_show_dead.get()):
				break
			elif reg_muta.status == 'equiv' and bool(self.var_show_equiv.get()):
				break
			else:
				cur -= 1
		self.var_mutant.set(str(cur))
		self.init_fields(cur)
		log(f"[Mutant] up -> {cur}")


	def on_op_equiv(self):
		log(f"[Operator.Equivalent] {bool(self.var_op_equiv.get())}")
		try:
			reg_muta = Mutant.get(Mutant.id==int(self.var_mutant.get()))
			if bool(self.var_op_equiv.get()):
				reg_muta.status = 'equiv'
			else:
				reg_muta.status = 'live'
			reg_muta.save()
		except:
			pass

	def on_show_alive(self):
		log(f"[Type to Show] Alive -> {bool(self.var_show_alive.get())}")

	def on_show_dead(self):
		log(f"[Type to Show] Dead -> {bool(self.var_show_dead.get())}")

	def on_show_equiv(self):
		log(f"[Type to Show] Equivalent -> {bool(self.var_show_equiv.get())}")


	def on_ok(self):
		log("[OK] clicado")

# ---------------------- Main ----------------------
def mutaview_gui(init_muta):
	root = tk.Tk()
#	style = ttk.Style()
	#print("Tema atual:", style.theme_use())
#	print("Temas disponíveis:", style.theme_names())
#	new_style = style.theme_names()[random.randint(0,len(style.theme_names()))]
#	style.theme_use(new_style)
#	print(new_style)
	root.title("View Mutants")
	# High-DPI / expansão
	root.rowconfigure(0, weight=1)
	root.columnconfigure(0, weight=1)
	app = ViewMutantsApp(root)
	root.minsize(900, 520)
	app.init_fields(init_muta)
	root.mainloop()






import ast
import pickle
from pyproteum.tcase import change_dir_connect
from pyproteum.models.models import *
from pyproteum.print_visit import PrintVisit
from pyproteum.mutaviewgui import mutaview_gui
import sys
from pyproteum.utiles import red,green



class MutaView(ast.NodeTransformer):
	
	def __init__(self, ast, function):
		self.ast = ast
		self.function = function
		self.source = None


	def go_visit(self):
		self.visit(self.ast) 
	
	
	def visit_FunctionDef(self, node):
		if node.name != self.function:
			self.generic_visit(node)
			return
		self.source = ast.unparse(node)


def get_text_from(text, li, lf):
	lines = text.split('\n')
	func =lines[li-1:lf]
	s = ''
	for r in func:
		s += r + '\n'
	return s

def __view():
	session_name = sys.argv[-1]
	directory = None
	muta_number = None
	list_number = None
	i = 2
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--x':
				i += 1
				list_number = sys.argv[i]
			case _:
				usage()
				return
		i += 1


	if not list_number:
		print('Mutant number not provided')
		sys.exit()

	try: 
		change_dir_connect(directory, session_name)
	except Exception as ex:
		print(red(f'Can not access test session'))
		print(red(ex))
		sys.exit()	
	
	for muta_number in list_number.split():
		try:
			muta_number = int(muta_number)
			reg_muta = Mutant.get(Mutant.id==muta_number)
			print('Mutant: ', reg_muta.id, reg_muta.operator, reg_muta.status)
		except Exception as ex:
			print(red(f'Can not access mutant number {muta_number}'))
			print(red(ex))
			sys.exit()

		try:
			s1,s2 = get_source_orig_muta(reg_muta)
			print(s1)
			print('------------------------')
			print(s2)
		except Exception as ex:
			print(red(f'Can not read mutant {reg_muta.id}'))
			print(red(ex))
			sys.exit()

	
def get_source_orig_muta(reg_muta, color=True):
	if reg_muta.function != '':
		f = open(reg_muta.source.filename)
		text_source = f.read().replace('\t', '    ')
		func_source = get_text_from(text_source, reg_muta.func_lineno, reg_muta.func_end_lineno)

		s1 = pp_format(func_source, -1, -1, reg_muta.func_lineno,color)
		astt = pickle.loads(reg_muta.ast)
		pv = PrintVisit(astt)
		pv.go_visit()
		func_source = get_text_from(str(pv), reg_muta.func_lineno, reg_muta.func_end_lineno)
		s2 = pp_format(func_source, reg_muta.lineno, reg_muta.end_lineno, reg_muta.func_lineno,color)
		f.close()
	else:
		f = open(reg_muta.source.filename)
		text_source = f.read().replace('\t', '    ')
		func_source = get_text_from(text_source, reg_muta.lineno, reg_muta.end_lineno)

		s1 =  pp_format(func_source, -1, -1, reg_muta.lineno,color)
		astt = pickle.loads(reg_muta.ast)
		pv = PrintVisit(astt)
		pv.go_visit()
		func_source = get_text_from(str(pv), reg_muta.lineno, reg_muta.end_lineno)
		s2 = pp_format(func_source, reg_muta.lineno, reg_muta.end_lineno, reg_muta.lineno,color)
		f.close()
	return (s1, s2)

def pp_format(src, begin, end, base,color):
	lines = src.split('\n')
	s = ''
	begin = begin - base
	end = end - base
	for k in range(len(lines)):
		if k >= begin and k <= end:
			if color:
				s +=  green('{:>4d}{}\t'.format(k+base,'-->') + lines[k] + '\n')
			else:
				s +=  '{:>4d}{}\t'.format(k+base,'-->') + lines[k] + '\n'
			
		else:
			s += '{:>4d} \t'.format(k+base) + lines[k] + '\n'
	return s



def __build():
	session_name = sys.argv[-1]
	directory = None
	muta_number = None
	i = 2
	list_number = None
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--x':
				i += 1
				list_number = sys.argv[i]
			case _:
				usage()
				return
		i += 1

	if not list_number:
		print('Mutant number not provided')
		sys.exit()

	try: 
		change_dir_connect(directory, session_name)
	except Exception as ex:
		print(red(f'Can not access test session'))
		print(red(ex))
		sys.exit()	
	
	for muta_number in list_number.split():
		try:
			muta_number = int(muta_number)
			reg_muta = Mutant.get(Mutant.id==muta_number)
			print('Mutant: ', reg_muta.id)
		except Exception as ex:
			print(red(f'Can not access mutant number {muta_number}'))
			print(red(ex))
			sys.exit()	
		
		muta_filename = '{}_muta{:04d}.py'.format(session_name,muta_number)
		astt = pickle.loads(reg_muta.ast)
		unv = PrintVisit(astt)
		unv.go_visit()
		try:
			f = open(muta_filename, 'w')
		#	tree = ast.parse(str(unv)) # verifica se código é válido.
			f.write(str(unv))
			print(f'File {muta_filename} succesfully created')
		except Exception as ex:
			print(red(f'Cannot create file {muta_filename}')	)
			print(red(ex))
		finally:
			f.close()

def __list():
	session_name = sys.argv[-1]
	directory = None
	i = 2
	while i < len(sys.argv[:-1]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case _:
				usage()
				return
		i += 1

	cont_dead = 0
	cont_live = 0
	cont_equiv = 0	
	try:
		change_dir_connect(directory, session_name)
		for muta in Mutant:
			print('{:04d} \t{}\t{}'.format(muta.id, muta.operator,muta.status))
			if muta.status == 'live':
				cont_live += 1
			elif muta.status == 'dead':
				cont_dead += 1
			else:
				cont_equiv += 1
	except Exception as ex:
		print(red(f'Can not access mutants'))
		print(red(ex))
		sys.exit()	
	print(green(f'Alive: {cont_live}'))
	print(green(f'Dead: {cont_dead}'))
	print(green(f'Equivalent: {cont_equiv}'))
	if cont_live+cont_dead == 0:
		print(green('Mutation score: {:.2f}'.format(0.0)))
	else:
		print(green('Mutation score: {:.3f}'.format(cont_dead/(cont_live+cont_dead))))	

def __gui():
	session_name = sys.argv[-1]
	directory = None
	muta_number = None
	i = 2
	list_number = None
	while i < len(sys.argv[:-2]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--x':
				i += 1
				list_number = sys.argv[i]
			case _:
				usage()
				return
		i += 1

	if not list_number:
		list_number = 1

	try: 
		change_dir_connect(directory, session_name)
	except Exception as ex:
		print(red(f'Can not access test session'))
		print(red(ex))
		sys.exit()	

	try:
		muta_number = int(list_number)
		reg_muta = Mutant.get(Mutant.id==muta_number)
	except Exception as ex:
		print(red(f'Can not access mutant number {muta_number}'))
		print(red(ex))
		sys.exit()	
	mutaview_gui(muta_number)

def main():
	n = len(sys.argv)-2
	if n < 1:
		usage()
	
	if sys.argv[1] == '--view':
		__view()
		return
	if sys.argv[1] == '--build':
		__build()
		return		
	if sys.argv[1] == '--list':
		__list()
		return	
	if sys.argv[1] == '--gui':
		__gui()
		return		
	else:
		usage()


def usage():
	print('Usage:')
	print('mutaview --view [--D <directory> ] --x <mutant number>  <session name>')
	print('\tShows original code and mutant code\n')
	print('mutaview --build [--D <directory> ] --x <mutant number>  <session name>')
	print('\tCreates a mutaNNNN.py file with the mutant code\n')
	print('mutaview --gui [--D <directory> ] [--x <mutant number>]  <session name>')
	print('\tShows original code and mutant code in a GUI. Allows to browser through all mutants\n')
	sys.exit()

if __name__ == '__main__' :
	main()



import ast
from pyproteum.moperators.myoperator import *


class Cccr(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
		self.const_set = list()

	def visit_FunctionDef(self, node):
		old_set = list(self.const_set)
		super().visit_FunctionDef(node)
		self.const_set = old_set
		return node

	def visit_Constant(self, node):
		if not id(node) in self.docstrings:
			old_value = node.value
			if not old_value in self.const_set:
				self.const_set.append(old_value)

			for new_value in self.const_set:
				if new_value == old_value:
					continue
				if type(new_value) != type(old_value):
					continue
				node.value = new_value
				self.salva_muta(node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1
				
			node.value = old_value
		self.generic_visit(node)
		return node
		



import ast
from pyproteum.moperators.myoperator import *

f = lambda x: str(x)

no_use = [ast.In(), ast.NotIn(), ast.Is(), ast.IsNot()]

use = [ast.Eq(), ast.NotEq(), ast.Lt(), ast.LtE(), ast.Gt(), ast.GtE()]

class Orrn(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.use = sorted(use, key=f)
		self.seq = 1
	
	def visit_Compare(self, node):
		for i,op in enumerate(node.ops):
			if self.check_in(op,no_use):
				#print('continue')
				continue

			self.ckeck_remove(op,self.use)
			
			for new_op in self.use:
				node.ops[i] = new_op
				#desloc = self.compute_len(op, new_op)
				self.salva_muta(node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1
			self.use = sorted(use, key=f)
			node.ops[i] = op
		
		return self.generic_visit(node)

	def check_in(self, op, use):
		op_class = type(op)
		it = list(filter(lambda x: isinstance(x,op_class), use))
		return len(it) > 0
		
	def ckeck_remove(self, op, use):
		for use_op in use:
			if type(op) is type(use_op):
				use.remove(use_op)
				return use_op
		return None

	def compute_len(self, op, new_op):
		r = 0
		if self.check_in(op, [ast.Eq(), ast.NotEq(), ast.LtE(), ast.GtE()]):
			if self.check_in(new_op, [ast.Lt(), ast.Gt()]):
				r = -1
		else:
			if self.check_in(new_op, [ast.Eq(), ast.NotEq(), ast.LtE(), ast.GtE()]):
				r = 1
		return r


import ast
from pyproteum.moperators.myoperator import *
from pyproteum.moperators.oaan import use

f = lambda x: str(x)


use = [ast.Add(),
ast.Sub(),
ast.Mult(),
ast.Div(),
ast.FloorDiv(),
ast.Mod(),
ast.Pow(),
ast.LShift(),
ast.RShift(),
ast.BitOr(),
ast.BitXor(),
ast.BitAnd(),
]

class Oaaa(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.use = sorted(use, key=f)
		self.seq = 1
	
	def visit_AugAssign(self, node):
		for i,op in enumerate(use):
			if type(op) == type(node.op):
				continue

			old_op = node.op
			node.op = op
			

			self.salva_muta(node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
			node.op = old_op
			self.seq += 1
		
		return self.generic_visit(node)



import ast
from pyproteum.moperators.myoperator import *
import copy


class Ccsr(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
		self.scalar_set = set()
		ParentSetter().visit(self.original)



	def visit_FunctionDef(self, node):
		old_set = set(self.scalar_set)
		super().visit_FunctionDef(node)
		self.scalar_set = old_set
		return node

	def visit_Name(self, node):
		if isinstance(node.ctx, ast.Store) :
			self.scalar_set.add(node.id)
		self.generic_visit(node)
		return node

	def visit_arg(self, node):
		if node.arg != 'self':
			self.scalar_set.add(node.arg)
		self.generic_visit(node)
		return node

	def visit_Constant(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

		if not id(node) in self.docstrings:
			for nv in sorted(list(self.scalar_set)):
				new_value = ast.Name(id=nv, ctx=ast.Load())
				ast.copy_location(new_value, node)
				if idx is not None:
					lst = getattr(p, fld)
					lst[idx] = new_value
				else:
					setattr(p,fld, new_value)
				self.salva_muta(new_value, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1

			if idx is not None: 
				lst = getattr(p, fld)
				lst[idx] = node
			else:
				setattr(p,fld, node)		

		self.generic_visit(node)
		return node
		



import ast
from pyproteum.moperators.myoperator import *
import copy


class Oodl(MyOperator):
	
	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1
		self.scalar_set = set()
		ParentSetter().visit(self.original)


	def visit_BinOp(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

		for new_value in [node.left, node.right]:
			if idx is not None:
				lst = getattr(p, fld)
				lst[idx] = new_value
			else:
				setattr(p,fld, new_value)
			self.salva_muta(p, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
			self.seq += 1

		if idx is not None: 
			lst = getattr(p, fld)
			lst[idx] = node
		else:
			setattr(p,fld, node)		

		self.generic_visit(node)
		return node

	def visit_UnaryOp(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

		for new_value in [node.operand]:
			if idx is not None:
				lst = getattr(p, fld)
				lst[idx] = new_value
			else:
				setattr(p,fld, new_value)
			self.salva_muta(p, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
			self.seq += 1

		if idx is not None: 
			lst = getattr(p, fld)
			lst[idx] = node
		else:
			setattr(p,fld, node)		

		self.generic_visit(node)
		return node

	def visit_BoolOp(self, node):
		p, fld, idx = node.parent, node.parent_field, node.parent_index

 #  nesse primeiro caso, o nó tem apenas 2 operandos. então retira o nó da árvore
		if len(node.values) == 2:			
			for new_value in node.values:
				if idx is not None:
					lst = getattr(p, fld)
					lst[idx] = new_value
				else:
					setattr(p,fld, new_value)
				self.salva_muta(p, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				self.seq += 1

			if idx is not None: 
				lst = getattr(p, fld)
				lst[idx] = node
			else:
				setattr(p,fld, node)		
		else: # nesse segundo caso, tira apenas um operando da árvore removendo da lista de operandos
			tam = len(node.values)
			for i in range(tam):
				old_node = node.values.pop(i)
				self.salva_muta(node, self.function, self.func_lineno, self.func_end_lineno,seq=self.seq)
				node.values.insert(i, old_node)
				self.seq += 1

		self.generic_visit(node)
		return node


import unittest
import os
import importlib

RED = '\033[31m'
GREEN = '\033[032m'
RESET = '\033[0m'

def red(s):
	if os.getenv('PYPROTEUMCOLOR') != None :
		return RED + str(s) + RESET
	return s 

def green(s):
	if os.getenv('PYPROTEUMCOLOR') != None :
		return GREEN + str(s) + RESET
	return s 



def get_test_files_with_discover(start_dir, pattern="test*.py"):
	loader = unittest.TestLoader()
	test_files = set()

	try:
		suite = loader.discover(start_dir, pattern=pattern)

	except:
		return test_files

	def _iter_suite(suite):
		for item in suite:
			if isinstance(item, unittest.TestSuite):
				_iter_suite(item)  # recursão direta
			else:
				module_name = item.__class__.__module__
				try:
					module = importlib.import_module(module_name)
					if hasattr(module, "__file__"):
							test_files.add(os.path.join(start_dir, os.path.basename(module.__file__)))  # só o nome curto
				except Exception:
					pass

	_iter_suite(suite)
	return test_files


import ast
from pyproteum.moperators.myoperator import *



class Scrb(MyOperator):
	


	def generic_visit(self, node):
		for _, value in ast.iter_fields(node):
			if isinstance(value, list):
				for k, item in enumerate(value):
					if isinstance(item,ast.Continue):
						new_node = ast.Break()
						ast.copy_location(new_node, item)
						value[k] = new_node
						self.salva_muta(new_node, self.function, self.func_lineno, self.func_end_lineno)
						value[k] = item

					self.visit(item)

			elif isinstance(value, ast.AST):
				self.visit(value)
		return node


from pyproteum import testnew, tcase, mutagen, exemuta, mutaview
import sys

if __name__ == '__main__':
	del sys.argv[0]
	if sys.argv ==[]:
		print('You should use one of the pyproteum commands.')
	else:	
		match sys.argv[0]: 
			case 'testnew':
				testnew.main()
			case 'tcase':
				tcase.main()
			case 'mutagen':
				mutagen.main()
			case 'exemuta':
				exemuta.main()
			case 'mutaview':
				mutaview.main()
			case _:
				print(f'Not found statement {sys.argv[0]}')




from pyproteum.moperators.orrn import Orrn
from pyproteum.moperators.ssdl import Ssdl
from pyproteum.moperators.oaan import Oaan
from pyproteum.moperators.oaaa import Oaaa
from pyproteum.moperators.sbrc import Sbrc
from pyproteum.moperators.scrb import Scrb
from pyproteum.moperators.crcr import Crcr
from pyproteum.moperators.cccr import Cccr
from pyproteum.moperators.ccsr import Ccsr
from pyproteum.moperators.oodl import Oodl
from pyproteum.moperators.oeap import Oeap
from pyproteum.moperators.oeap import Oepa


def get_descr(name):
    for r in operator_list:
        if r['name'] == name:
            return r['descr']
    return None

operator_list = [
	{'class': Cccr, 
    'descr':'Cosntant by Constant Replacement',
	'name' : 'cccr'},

	{'class': Ccsr, 
    'descr':'Constant by scalar Replacement',
	'name' : 'ccsr'},
    
	{'class': Crcr, 
    'descr':'Required Constant Replacement',
	'name' : 'crcr'},
    
	{'class':Oaaa,
	'descr': 'Replace arithmetic assigment operator by other arithmetic assigment operator',
	'name':'oaaa'},

	{'class':Oaan,
	'descr': 'Replace arithmetic operator by other arithmetic operator',
	'name':'oaan'},

	{'class':Oeap,
	'descr': 'Replace augmented assignment by plain assignment',
	'name':'oeap'},
    
	# {'class':Oepa,
	# 'descr': 'Replace plain assignment by augmented assignment',
	# 'name':'oepa'},
    
	{'class':Oodl,
	'descr': 'Operator deletion',
	'name':'oodl'},

	{'class': Orrn, 
    'descr':'Replace relational operator (< > <= >= == !=) by other relational operator',
	'name' : 'orrn'},

	{'class':Sbrc,
	'descr': 'Replace each break statement by a continue statement',
	'name':'sbrc'},

	{'class':Scrb,
	'descr': 'Replace each continue statement by a break statement',
	'name':'scrb'},

	{'class':Ssdl,
	'descr': 'Replace each statement by a pass statement',
	'name':'ssdl'},
	]




import ast
from pyproteum.moperators.myoperator import *



class Ssdl(MyOperator):
	

	def __init__(self, original,filename):
		super().__init__(original, filename)
		self.seq = 1

	def generic_visit(self, node):		
		for campo, value in ast.iter_fields(node):
			if isinstance(value, list):
				for k, item in enumerate(value):
					if id(item) in self.docstrings:
						self.visit(item)								
					elif isinstance(item, ast.If): #trata o caso especial do elif
						pass_node = ast.Pass()
						if campo == 'orelse' and len(value) == 1 and item.col_offset == node.col_offset:
							firstc = item.body[0]
							ast.copy_location(pass_node, firstc)
						else:
							ast.copy_location(pass_node, item)
						value[k] = pass_node
						# self.show_node(pass_node)
						# self.show_node(item)
						self.salva_muta(pass_node, self.function, self.func_lineno, self.func_end_lineno, self.seq)
						value[k] = item				
					elif self._is_command(item):
						pass_node = ast.Pass()
						ast.copy_location(pass_node, item)
						value[k] = pass_node
						# self.show_node(pass_node)
						# self.show_node(item)
						self.salva_muta(pass_node, self.function, self.func_lineno, self.func_end_lineno, self.seq)
						value[k] = item
					self.seq += 1
					self.visit(item)


			elif isinstance(value, ast.AST):
				self.visit(value)
		return node

	def _is_command(self, node):
		return isinstance(node, (
			ast.If, ast.For, ast.AsyncFor, ast.While, ast.With, ast.Try,
			ast.Assign, ast.AugAssign, ast.Expr, ast.AsyncWith,
			ast.Return, ast.Raise, ast.Assert, ast.Delete, ast.Nonlocal,
			ast.Break, ast.Continue, ast.Match, ast.AnnAssign, ast.Global
		))


import argparse
import os, sys
from pyproteum.models.models import *
import pickle
import ast
from pyproteum.utiles import red,green

def main():
	parser = argparse.ArgumentParser(description="Create a new test session for a Python program")


	# Flags exclusivas --test ou --research
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--test", action="store_true", help="Create a regular test session (default)")
	group.add_argument("--research", action="store_true", help="Create a research session")

	# Parâmetros opcionais
	parser.add_argument("--S", dest="source_files", action="append", help="Name of a source file; use multiple --S for multiple files")
	parser.add_argument("--D", dest="directory", help="Directory where the session will be created")

	# Nome da sessão de teste (obrigatório e posicional)
	parser.add_argument("session_name", help="Name of the test session")

	args = parser.parse_args()

	# Atribui valores padrão se não fornecidos
	source_files = args.source_files or [args.session_name+'.py']
	if args.directory is None or args.directory == '.':
		directory = os.getcwd()
	else:
		directory = args.directory
		
	session_type = "research" if args.research else "test"

	# Saída (poderia ser substituída pela criação real dos arquivos)
	print(f"Creating a {session_type} test session...")
	print(f"Session name: {args.session_name}")
	print(f"Source files: {', '.join(sf for sf in source_files)}")
	print(f"Directory: {directory}")



	dbname = args.session_name+'.db'
	try: 
		if directory:
			os.chdir(directory)
		if os.path.exists(dbname):
			os.remove(dbname)
		database = SqliteDatabase(dbname)
		db.initialize(database)  # aqui o proxy é vinculado ao banco real
		db.pragma('foreign_keys', 1, permanent=True)
	except Exception as ex:
		print(red(f'Can not find database {dbname}'))
		print(red(ex))
		sys.exit()


	database.drop_tables([Session, TestCase, Mutant,])
	database.create_tables([Session, TestCase, Mutant])
	for srcname in source_files:
		try: 
			f = open(srcname)
			src = f.read().replace('\t', '    ')
			axt = ast.parse(src)
		except FileNotFoundError as ex:
			print(red(f'Can not use source file: {srcname}'))
			print(red(ex))
			sys.exit()			
		except Exception as ex:
			print(red(f'Can not use source file: {srcname}'))
			print(red(ex))
			f.close()
			sys.exit()			
		
		Session.create( filename=srcname,
			type = session_type,
			) 
	

	print("Test session created successfully.")

if __name__ == "__main__":
	main()



from setuptools import setup, find_packages

setup(
	name='pyproteum',
	version='0.1',
	packages=find_packages(),
	install_requires=[
		'timeout-decorator',
		'peewee'
	],
	author='Delamaro',
	description='A tool for mutation testing in Python. Implemented as a Python module.',
	python_requires='>=3.10',
	# Note: 'pickle' is part of the Python standard library and should not be listed in install_requires.
)



import ast, os, sys
import copy
from pyproteum.print_visit import PrintVisit
from pyproteum.utiles import red,green


class MyOperator(ast.NodeTransformer):
	
	def __init__(self, original,filename):
		self.original = copy.deepcopy(original)
		self.not_changr_original = original
		self.mutants = []
		self.function = ''
		self.func_lineno = 0
		self.func_end_lineno = 0		
		self.docstrings = set()
		self.pre_visit(self.original)
		self.filename = filename

	def pre_visit(self, tree):
		for node in ast.walk(tree):
			try:
				if ast.get_docstring(node):
					fn = node.body[0]
					self.docstrings.add(id(fn))
			except: 
				continue

	def salva_muta(self, node, func, ln, eln, seq=1):
		cop = copy.deepcopy(self.original)
		ast.fix_missing_locations(cop)
		from pyproteum.exemuta import create_module_from_ast

		try: 
			nll = open(os.devnull, 'w')
			err = sys.stderr
			out = sys.stdout
			sys.stdout = sys.stderr = nll
			create_module_from_ast(self.filename, cop)
		except Exception as ex:
			sys.stdout = out
			sys.stderr = err			
			print(red(f'Skiping mutant {str(self)}'))
			#print(ex)
			return
		finally:
			nll.close()
			sys.stderr = err
			sys.stdout = out
		
		if self.check_sintaxe(cop):
			muta = MutantDict({'function': func,
					'func_lineno': ln,
					'func_end_lineno': eln,
					'operator': str(self),
					'lineno':node.lineno,
					'col_offset':node.col_offset,
					'end_lineno':node.end_lineno,
					'end_col_offset':node.end_col_offset,
					'seq_number' : seq,
					'ast':cop
					})
			self.mutants.append(muta)
		else:
			print(red(f'Skiping mutant {str(self)}'))
		


		
	def go_visit(self):
		self.visit(self.original)

	def visit_FunctionDef(self, node):
		old_f = self.function
		old_l = self.func_lineno
		old_el = self.func_end_lineno
		self.function = node.name
		self.func_lineno = node.lineno
		self.func_end_lineno = node.end_lineno
		r = self.generic_visit(node)
		self.function = old_f
		self.func_end_lineno = old_el
		self.func_lineno = old_l
		return r

	def visit_Expr(self, node):
		if id(node) in self.docstrings:
			return node		
		return self.generic_visit(node)

	def __str__(self):
		s = str(type(self))
		return s [-6:-2].lower()
		

	def check_sintaxe(self,tree):
		unv = PrintVisit(tree)
		try:
			unv.go_visit()
			ast.parse(str(unv))
			return True
		except Exception as ex:
			#print(unv)
			#print(ex)
			#datv = DotASTVisitor()
			#datv.visit(tree)
			#datv.write('dot.dot')
			#datv = DotASTVisitor()
			#datv.visit(self.not_changr_original)
			#datv.write('dot2.dot')			
			return False

	
	def show_node(self, node):
		print('\n', node)
		if hasattr(node, 'lineno'):
			print(f'lineno: {node.lineno}')			
			print(f'end_lineno: {node.end_lineno}')
			print(f'col_offset: {node.col_offset}')	
			print(f'end_col_offset: {node.end_col_offset}')					

class MutantDict(dict):


	def __lt__(self, other):
		if self['operator'] < other['operator']:
			return True
		if self['operator'] > other['operator']:
			return False
		if self['lineno'] < other['lineno']:
			return True
		if self['lineno'] > other['lineno']:
			return False
		if self['col_offset'] < other['col_offset']:
			return True
		if self['col_offset'] > other['col_offset']:
			return False
		if self['seq_number'] < other['seq_number']:
			return True
		if self['seq_number'] > other['seq_number']:
			return False
		return False


# This class is used to set the parent of each node.
# Not all operators use this
# CCsr 
class ParentSetter(ast.NodeVisitor):
    def visit(self, node):
        # garante que todo nó tenha attrs (pai/campo/índice podem já ter sido definidos)
        if not hasattr(node, "parent"):
            node.parent = None
            node.parent_field = None
            node.parent_index = None
        super().visit(node)

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, ast.AST):
                value.parent = node
                value.parent_field = field
                value.parent_index = None
                self.visit(value)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, ast.AST):
                        item.parent = node
                        item.parent_field = field
                        item.parent_index = i
                        self.visit(item)


class DotASTVisitor(ast.NodeVisitor):
    def __init__(self):
        self.output = ["digraph ast {"]
        self.node_id = 0
        self.stack = []

    def _gen_id(self):
        self.node_id += 1
        return f"n{self.node_id}"

    def _format_label(self, node):
        # Começa com o nome da classe
        label = type(node).__name__
        if hasattr(node, 'lineno'):
            label += f'\\nline:=({node.lineno},{node.col_offset})'

        # Enriquecer com informação relevante
        if isinstance(node, ast.Name):
            label += f'\\nid={node.id}'
        elif isinstance(node, ast.Constant):
            label += f'\\nvalue={repr(node.value)}'
        elif isinstance(node, ast.arg):
            label += f'\\narg={node.arg}'
        elif isinstance(node, ast.FunctionDef):
            label += f'\\nname={node.name}'
        elif isinstance(node, ast.ClassDef):
            label += f'\\nname={node.name}'
        elif isinstance(node, ast.Attribute):
            label += f'\\nattr={node.attr}'

        return label

    def visit(self, node):
        cur_id = self._gen_id()
        label = self._format_label(node)
        self.output.append(f'{cur_id} [label="{label}"];')

        if self.stack:
            parent_id = self.stack[-1]
            self.output.append(f"{parent_id} -> {cur_id};")

        self.stack.append(cur_id)
        super().visit(node)
        self.stack.pop()

    def get_dot(self):
        return "\n".join(self.output) + "\n}"

    def write(self, arq):
        with open(arq,'w') as f:
            f.write(self.get_dot()+'\n')


import types, sys, ast, multiprocessing
import sys, csv
import os,io
from pyproteum.models.models import *
import unittest
import pickle
import types
from pyproteum.moperators.myoperator import *
from pyproteum.tcase import change_dir_connect
from pyproteum.tcase import load_module, get_all_test_names, _ensure_module_hierarchy
from pyproteum.utiles import red,green
import timeout_decorator
import inspect


class RegistrandoResultado(unittest.TextTestResult):
	def __init__(self, stream, descriptions, verbosity, number, research):
		super().__init__(stream, descriptions, verbosity)
		self.my_successes = []  # armazena os testes que passaram
		self.timeouts = []
		self.all = []		# todos os testes executados
		self.all_order = {}
		self.order = number
		self.count = 0
		self.research = research

	def startTest(self, test):
		self.all.append(test)
		self.count += 1
		self.all_order[test.id()] = '{}:{}'.format(self.order, self.count)
		super().startTest(test)

	def addSuccess(self, test):
		self.my_successes.append(test)
		super().addSuccess(test)

	def stopTest(self, test):
		super().stopTest(test)
	
	def addFailure(self, test, err):
		etype, evalue, tb = err
		# Marcação explícita de TIMEOUT
		if getattr(etype, "__name__", "") in ("TimeoutError", "TimeoutError_", "TimeoutException"):
			# opcional: anotar em um campo extra
			self.timeouts.append(test)				
		super().addFailure(test, err)
		if not self.research:
			self.stop()

	def addError(self, test, err):
		etype, evalue, tb = err
		# Marcação explícita de TIMEOUT
		if getattr(etype, "__name__", "") in ("TimeoutError", "TimeoutError_", "TimeoutException"):
			# opcional: anotar em um campo extra
			self.timeouts.append(test)		
		super().addError(test, err)
		if not self.research:
			self.stop()


class RegistrandoRunner(unittest.TextTestRunner):

	def __init__(self, *args, number=0, research=False, **kwargs):
		super().__init__(*args, **kwargs)
		self.number = number
		self.research = research

	def _makeResult(self):
		return RegistrandoResultado(self.stream, self.descriptions, self.verbosity, self.number,self.research)




def create_module_from_astx(nome_modulo, arvore_ast):
	modulo = types.ModuleType(nome_modulo)
	code = compile(arvore_ast, filename="<ast>", mode="exec")
	exec(code, modulo.__dict__)
	sys.modules[nome_modulo] = modulo
	return modulo

import types, sys, multiprocessing, os, io
from contextlib import redirect_stdout

def _check_ast_executesx(code_obj, nome, suppress_output):
	# Executa apenas para ver se trava
	modulo = types.ModuleType(nome)

	if suppress_output:
		with open(os.devnull, 'w') as fnull:
			out = sys.stdout
			err = sys.stderr
			sys.stdout = stderr = fnull
			exec(code_obj, modulo.__dict__)
			sys.stdout = out
			sys.stderr = err
	else:
		exec(code_obj, modulo.__dict__)

def create_module_from_astxx(nome_modulo, arvore_ast, timeout=3, suppress_output=True):
	code = compile(arvore_ast, filename="<ast>", mode="exec")

	# Executa o código compilado em um subprocesso apenas para validar
	p = multiprocessing.Process(target=_check_ast_executes, args=(code, nome_modulo, suppress_output))
	p.start()
	p.join(timeout)

	if p.is_alive():
		p.terminate()
		p.join()
		raise TimeoutError(f"O módulo '{nome_modulo}' não respondeu em {timeout} segundos.")

	# Se passou no teste, agora podemos executar de verdade no processo principal
	modulo = types.ModuleType(nome_modulo)

	if suppress_output:
		with open(os.devnull, 'w') as fnull:
			out = sys.stdout
			err = sys.stderr
			sys.stdout = stderr = fnull
			exec(code, modulo.__dict__)
			sys.stdout = out
			sys.stderr = err
	else:
		exec(code, modulo.__dict__)

	sys.modules[nome_modulo] = modulo
	return modulo




def _check_ast_executes(code_obj, path_filename, suppress_output):
	module_name, parent, parts = _ensure_module_hierarchy(path_filename)

	modulo = types.ModuleType(module_name)
	if suppress_output:
		with open(os.devnull, 'w') as fnull:
			old_out, old_err = sys.stdout, sys.stderr
			sys.stdout, sys.stderr = fnull, fnull
			try:
				exec(code_obj, modulo.__dict__)
			finally:
				sys.stdout, sys.stderr = old_out, old_err
	else:
		exec(code_obj, modulo.__dict__)

	sys.modules[module_name] = modulo
	if parent is not None:
		setattr(parent, parts[-1], modulo)

def create_module_from_ast(path_filename, arvore_ast, timeout=3, suppress_output=True):
	code = compile(arvore_ast, filename="<ast>", mode="exec")

	# Testa em subprocesso com timeout
	p = multiprocessing.Process(target=_check_ast_executes, args=(code, path_filename, suppress_output))
	p.start()
	p.join(timeout)
	if p.is_alive():
		p.terminate()
		p.join()
		raise TimeoutError(f"O módulo '{path_filename}' não respondeu em {timeout} segundos.")

	# Executa de verdade no processo principal
	module_name, parent, parts = _ensure_module_hierarchy(path_filename)

	modulo = types.ModuleType(module_name)
	if suppress_output:
		with open(os.devnull, 'w') as fnull:
			old_out, old_err = sys.stdout, sys.stderr
			sys.stdout, sys.stderr = fnull, fnull
			try:
				exec(code, modulo.__dict__)
			finally:
				sys.stdout, sys.stderr = old_out, old_err
	else:
		exec(code, modulo.__dict__)

	sys.modules[module_name] = modulo
	if parent is not None:
		setattr(parent, parts[-1], modulo)

	return modulo




def __exec():
	session_name = sys.argv[-1]
	directory = None
	i = 2
	keep = False
	verbose = False
	while i < len(sys.argv[:-1]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--keep':
				keep = True
			case '--v':
				verbose = True
			case _:
				usage()
				return
		i += 1
	
	change_dir_connect(directory, session_name)
	session = Session.get(Session.id==1)
	rse = session.type == 'research'

	fields = get_all_test_names()
	Execution = create_exec_db(fields)

	cont_dead = 0
	cont_live = 0
	cont_equiv = 0

	with db.atomic():
		for muta in Mutant:
			reg_muta = Execution.create(mutant=muta)

			if muta.status == 'equiv':
				cont_equiv += 1
				reg_muta.save()
				continue

			if keep and muta.status != 'live' and not rse:
				cont_dead += 1
				reg_muta.save()
				continue

			print('Mutant {} -- {}'.format(muta.id, muta.operator))
			tree = pickle.loads(muta.ast)
			create_module_from_ast(muta.source.filename, tree)

			#print(reg_muta)
			dead = False
			for testfile in TestCase:
				try: 
					nll = open(os.devnull, 'w')
					err = sys.stderr
					out = sys.stdout
					sys.stdout = sys.stderr = nll					
					module_test = load_module(testfile.filename)
					aplicar_timeout_em_tests(module_test)
					suite = unittest.TestLoader().loadTestsFromModule(module_test)
					resultado = RegistrandoRunner(stream=io.StringIO(), research=rse,verbosity=0,number=testfile.id).run(suite)
					sys.stdout = out
					sys.stderr = err
				except Exception as ex:
					sys.stdout = out
					sys.stderr = err					
					print(red(f'Error. Can not run test file {testfile}'))
					print(red(ex))
					sys.exit()
				
				for test in resultado.my_successes:
					t = resultado.all_order[test.id()]
					if verbose:
						print(f'{t} passed.')
					setattr(reg_muta, test.id(), 'live')
					reg_muta.save()
					
				for test,_ in resultado.failures:
					t = resultado.all_order[test.id()]
					if test in resultado.timeouts:
						s = 'timeout'
					else:
						s = 'fail'
					if verbose:
						print(f'{t} {s}.')
					
					setattr(reg_muta, test.id(), s)
					reg_muta.save()				
					dead = True
					

				for test,_ in resultado.errors:
					t = resultado.all_order[test.id()]
					if verbose:
						print(f'{t} error.')
					setattr(reg_muta, test.id(), 'error')
					reg_muta.save()
					dead = True
				if dead and not rse:
					break

					
			if dead:
				muta.status = 'dead'
				print('Dead')
				cont_dead += 1
				muta.save()
			else:
				cont_live += 1
				muta.status = 'live'
				muta.save()
	
	print(green(f'Alive: {cont_live}'))
	print(green(f'Dead: {cont_dead}'))
	print(green(f'Equivalent: {cont_equiv}'))
	if cont_live+cont_dead == 0:
		print(green('Mutation score: {:.2f}'.format(0.0)))
	else:
		print(green('Mutation score: {:.3f}'.format(cont_dead/(cont_live+cont_dead))))

	
def create_exec_db(fields):
	model_dict = { 'mutant' : ForeignKeyField(Mutant, backref='executed')}
	fields = sorted(fields)
	for r in fields:
		model_dict[r] = TextField(default='No Exec')	
	Execution = criar_modelo('Execution', model_dict)
	db.drop_tables([Execution])
	db.create_tables([Execution])
	return Execution

TIMEOUT_SEGUNDOS = 2

def aplicar_timeout_em_tests(modulo):
	for nome, obj in inspect.getmembers(modulo):
		if inspect.isclass(obj) and issubclass(obj, unittest.TestCase):
			for metodo_nome, metodo in inspect.getmembers(obj, inspect.isfunction):
				if metodo_nome.startswith("test_"):
					setattr(obj, metodo_nome, timeout_decorator.timeout(TIMEOUT_SEGUNDOS, use_signals=False)(metodo))
	
def __csv():
	session_name = sys.argv.pop()
	directory = None
	outname = None
	i = 2
	while i < len(sys.argv[:-1]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--O':
				i += 1
				outname = sys.argv[i]
			case _:
				usage()
				return
		i += 1

	change_dir_connect(directory, session_name)

	if outname is None:
		outname = session_name+'.csv'

	try:
		Execution = get_table_model('Execution')
	except Exception as ex:
		print(red('Previous execution not found. Try "exemuta.py --exec" before exporting'))
		print(red(ex))
		sys.exit()

	try:
		with open(outname, 'w') as out:
			writer = csv.writer(out)
			campos = [field.name for field in Execution._meta.sorted_fields]
			campos.remove('id')
			writer.writerow(campos)
			for registro in Execution.select():
				writer.writerow([getattr(registro, campo) for campo in campos])
		print(f'{outname} successfully generated')
	except Exception as ex:
		print(red(f'Could not generate {outname}'))
		print(red(ex))
		sys.exit()


def __equiv():
	session_name = sys.argv[-1]
	directory = None
	list_number = None
	i = 2
	while i < len(sys.argv[:-1]):
		s = sys.argv[i]
		match s:
			case '--D':
				i += 1
				directory = sys.argv[i]
			case '--x':
				i += 1
				list_number = sys.argv[i]
			case _:
				usage()
				return
		i += 1


	try: 
		change_dir_connect(directory, session_name)
	except Exception as ex:
		print(red(f'Can not access test session'))
		print(red(ex))
		sys.exit()	

	# se os número não foram fornecidos, marca todos os vivos
	if  list_number is None:
		list_number = [] 
		for numero in Mutant.select().where(Mutant.status == "live"):
			list_number.append(numero.id)
	else:
		list_number = list_number.split()

	with db.atomic():
		for muta_number in list_number:
			try:
				muta_number = int(muta_number)
				reg_muta = Mutant.get(Mutant.id==muta_number)
				print('Mutant: ', reg_muta.id)
				if not reg_muta.status in ['live','equiv']:
					print(f'Warning: mutant {reg_muta.id} is not alive. It is {reg_muta.status}')
				else:
					reg_muta.status = 'equiv'
					reg_muta.save()
			except Exception as ex:
				print(red(f'Can not access mutant number {muta_number}'))
				print(red(ex))
				#sys.exit()

def main():
	n = len(sys.argv)-2
	if n < 1:
		usage()
	
	if sys.argv[1] == '--exec':
		__exec()
		return
	elif sys.argv[1] == '--csv':
		__csv()
		return
	elif sys.argv[1] == '--equiv':
		__equiv()
		return
	else:
		usage()


def usage():
	print('Usage:')
	print('exemuta --exec [--keep] [--D <directory> ]  <session name>')
	print('\tExecute the mutants with the test cases in the session')
	print('\t--keep: execute only the live mutants')
	print('exemuta --csv [--D <directory> ]  <session name>')
	print('\tExports the last execution to a csv file')
	print('exemuta --equiv --x <list of numbers> [--D <directory> ]  <session name>')
	print('\tMarks mutants as equivalents') 
	sys.exit()

if __name__ == '__main__' :
	main()
